# Copyright 2016-2018 Carl Poelking, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from builtins import range
from builtins import object
import numpy as np
import math
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.caching_backend import CachingLevel
import re, os, sys, json, logging


class DL_POLYParserContext(object):
	def __init__(self):
		self.cell	= []
		return
	def onClose_section_run(self, backend, gIndex, section):
		print("<onClose_section_run>")
		return
	def onClose_section_method(self, backend, gIndex, section):
		print("<onClose_section_method>")
		return
	def onClose_section_system(self, backend, gIndex, section):
		print("<onClose_section_system>")
		return
	def onClose_x_dl_poly_section_md_molecule_type(self, backend, gIndex, section):
		print("<onClose_molecule_type>")
		return

def test_adhoc(parser):
	return None
	for i in range(10):
		ln = parser.fIn.readline()
		print(ln)
		parser.fIn.pushbackLine(ln)
	#print len(parser.backend.openSections)
	return None

mainFileDescription = SM(name = 'root',
	 weak = True,
	 startReStr = "",
	 subMatchers = [

		SM(name = 'newRun',
		    startReStr = r" \*\* DL_POLY \*\*  authors:   i.t.todorov   &   w.smith  \** P \**",
		    repeats = False,
		    required = True,
		    forwardMatch = True,
			adHoc = test_adhoc,
		    sections   = ['section_run'],
		    subMatchers = [
				SM(name = 'progHeader',
					startReStr = r" \*\* DL_POLY \*\*  authors:   i.t.todorov   &   w.smith  \** P \**",
					subMatchers = [
						SM(r"\s*\*\*\s*\*\*  version:  (?P<program_version>[0-9a-zA-Z_.]*)    / \s* (?P<x_dl_poly_program_version_date>[\w]*\s*[\w]*)\s*\** O \**"),
						SM(r"\s*\**  when publishing research data obtained using (?P<program_name>[0-9a-zA-Z_.]*)  \**"),
						SM(r"\s*\**\s*(?P<x_dl_poly_system_description>[0-9a-zA-Z_()]*)\s*\**")
					]),
				SM(name = 'mdParams',
					startReStr = r"\s*SIMULATION CONTROL PARAMETERS\s*",
					required = True,
					sections = ['section_method'],
					subMatchers = [
						SM(r"\s*simulation temperature \(K\)\s*(?P<x_dl_poly_thermostat_temperature>[-+0-9.eEdD]*)\s*"),
						SM(r"\s*equilibration period \(steps\)\s*(?P<x_dl_poly_step_number_equilibration>[0-9]*)\s*"),
						SM(r"\s*selected number of timesteps\s*(?P<x_dl_poly_step_number>[0-9]*)\s*")
					]),

				# Open system
				# ... open topology
				# ... ... store <numberofmoleculetypes>
				# ... ... open molecule type id1
				# ... ... open molecule type ...

				SM(name = 'mdSystem',
					startReStr = r"\s*SYSTEM SPECIFICATION\s*",
					required = True,
					sections = ['section_system'],
					subMatchers = [
						SM(name = 'mdTopology',
							startReStr = r"\s*number of molecular types\s*[0-9]*\s*",
							required = True,
							forwardMatch = True,
							sections = ['x_dl_poly_section_md_topology'],
							subMatchers = [
								SM(r"\s*number of molecular types\s*(?P<x_dl_poly_md_molecular_types>[0-9]*)"),
								SM(name = 'mdTopologyMolecule',
									startReStr = r"\s*molecular species type\s*[0-9]*\s*",
									repeats = True,
									required = True,
									forwardMatch = True,
									sections = ['x_dl_poly_section_md_molecule_type'],
									subMatchers = [
										SM(r"\s*molecular species type\s*(?P<x_dl_poly_md_molecule_type_id>[0-9]*)\s*"),
										SM(r"\s*name of species:\s*(?P<x_dl_poly_md_molecule_type_name>[\w]* [\w]*)\s*")
									])
							])
						#SM(r"\s*number of molecular types\s*(?P<dl_poly_molecular_types_number>[0-9]*)\s*"),
						#SM(name = 'mdTopologyMolecule',
						#startReStr = r"\s*molecular species type\s*[0-9]*\s*",
						#repeats = True,
						#required = True,
						#sections = ['dl_poly_section_molecule_types'],
						#subMatchers = [\
						#])
					])
			])
	])


"""
mainFileDescription = SM(name = 'root',
     weak = True,
     startReStr = "",
     subMatchers = [

        SM(name = "newRun",
        startReStr = r"",
        #repeats = True,
        required = True,
        forwardMatch = True,
        sections   = ["section_run"],
        subMatchers = [

            SM(startReStr = "Number of k-points\s*",
               forwardMatch = True,
               sections = ["section_system"],
               subMatchers = [

                   SM(startReStr = "Number of k-points\s*",
                      forwardMatch = True,
                      sections = ["section_single_configuration_calculation"],
                      subMatchers = [

                          SM(startReStr = "Number of k-points\s*",
                             forwardMatch = True,
                             sections = ["section_eigenvalues_group"],
                             subMatchers = [

                                 SM(startReStr = "Number of k-points\s*",
                                    sections = ["section_eigenvalues"],
                                    forwardMatch = True,
                                    subMatchers = [

                                     SM(startReStr = "K\-point\s*[0-9]+\s*",
                                        sections = ["dl_poly_section_k_points"],
                                        forwardMatch = True,
                                        repeats = True,
                                        subMatchers = [

                                            SM(r"K\-point\s*[0-9]+\s*(?P<dl_poly_store_k_points> [-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                                               repeats = True),

                                                SM(name = 'Eigen',
                                                   startReStr = r"Spin component\s*[0-9]+\s*",
                                                   sections = ['dl_poly_section_eigenvalues'],
                                                   repeats = True,
                                                   subMatchers = [

                                                      SM(r"\s*(?P<dl_poly_store_eigenvalues> [-\d\.]+)",
                                                         repeats = True)

                                                   ]), # CLOSING dl_poly_section_eigenvalues

                                        ]), # CLOSING dl_poly_section_k_points

                                    ]), # CLOSING section_eigenvalues

                            ]), # CLOSING section_eigenvalues_group

                      ]), # CLOSING section_single_configuration_calculation

            ]), # CLOSING section_system

]), # CLOSING SM newRun


])
"""

# THIS PARSER
parserInfo = {'name':'dl_poly-parser', 'version': '0.0'}

# # LOAD METADATA
# metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/dl_poly.nomadmetainfo.json"))
# metaInfoEnv, warnings = loadJsonFile(filePath = metaInfoPath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)

# CUSTOMIZE CACHING
cachingLevelForMetaName = {}


class DlPolyParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       from unittest.mock import patch
       logging.info('dl-poly parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("dl_poly.nomadmetainfo.json")
       with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
           mainFunction(
               mainFileDescription=mainFileDescription,
               metaInfoEnv=None,
               parserInfo = parserInfo,
               cachingLevelForMetaName = cachingLevelForMetaName,
               superContext=DL_POLYParserContext(),
               superBackend=backend,
			   onClose = {},
               defaultSectionCachingLevel = True)

       return backend



# Copyright 2016-2018 Fawzi Mohamed, Carl Poelking, Daria Tomecka
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import range
import os
import sys
import re
import json
import logging
# import setup_paths
import numpy as np

from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from contextlib import contextmanager

from libatomsparser.libLibAtomsParser import *

# try:
#     from libMomo import osio, endl, flush
#     osio.ConnectToFile('parser.osio.log')
#     green = osio.mg
# except:
osio = endl = flush = None
green = None

parser_info = {
    "name": "parser-lib-atoms",
    "version": "0.0",
    "json": "../../../../nomad-meta-info/meta_info/nomad_meta_info/lib_atoms.nomadmetainfo.json"
}


class LibAtomsParserWrapper():
    """ A proper class envolop for running this parser using Noamd-FAIRD infra. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        logging.info('lib-atoms parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("lib_atoms.nomadmetainfo.json")
        # Call the old parser without a class.
        parserInfo = {'name': 'lib_atoms-parser', 'version': '0.0'}
        backend = parse_without_class(mainfile, backend, parserInfo)
        return backend


# LOGGING
def log(msg, highlight=None, enter=endl):
    # if osio:
    #     if highlight==None: hightlight = osio.ww
    #     osio << highlight << msg << enter
    logging.debug(msg)
    return

# CONTEXT GUARD
@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)

def push(jbe, terminal, key1, fct=lambda x: x.As(), key2=None):
    if key2 == None: key2 = key1
    value =  fct(terminal[key2])
    jbe.addValue(key1, value)
    return value

def push_array(jbe, terminal, key1, fct=lambda x: x.As(), key2=None):
    if key2 == None: key2 = key1
    value =  np.asarray(fct(terminal[key2]))
    jbe.addArrayValues(key1, value)
    return value

def push_value(jbe, value, key):
    jbe.addValue(key, value)
    return value

def push_array_values(jbe, value, key):
    jbe.addArrayValues(key, value)
    return value

def parse_without_class(output_file_name, backend, parser_info):
    """ Parse method to parse mainfile and write output to backend."""
    jbe = backend
    jbe.startedParsingSession(output_file_name, parser_info)
    base_dir = os.path.dirname(os.path.abspath(output_file_name))

    # jbe.startedParsingSession(output_file_name, parser_info)

    # base_dir = os.path.dirname(os.path.abspath(output_file_name))
    terminal_gap = LibAtomsGapParser(osio)
    terminal_gap.ParseOutput(output_file_name, base_dir)
    terminal_trj = terminal_gap.trj

    logging.debug("Start parsing...")
    # osio << "Start parsing ..." << osio.endl
    logging.debug('Base directory = %s' % base_dir)
    # osio << "Base directory = '%s'" % base_dir << osio.endl

    gap = terminal_gap
    trj = terminal_trj

    length_libatoms2nomad = 1e-10 # angstrom to meter
    energy_libatoms2nomad = 1.60217662e-19

    with open_section(jbe, 'section_run') as gid_run:
        push(jbe, gap, 'program_name')
        push(jbe, gap, 'program_version', key2='GAP_params.svn_version')

        # SYSTEM DESCRIPTION
        refs_system_description = []
        all_frames = trj.frames # <- Initial config + trajectory
        for frame in all_frames:
            with open_section(jbe, 'section_system') as gid:
                refs_system_description.append(gid)
                # Configuration core
                atom_labels = np.array(frame.ase_config.get_chemical_symbols())
                push_array_values(jbe, atom_labels, 'atom_labels')
                push_array_values(jbe, frame.ase_config.get_positions()*length_libatoms2nomad, 'atom_positions')
                # CAREFUL, OBSERVE CELL CONVENTION: 1st index -> x,y,z, 2nd index -> a,b,c
                # Hence use transpose here:
                push_array_values(jbe, frame.ase_config.get_cell().T*length_libatoms2nomad, 'simulation_cell')
                push_array_values(jbe, frame.ase_config.get_pbc(), 'configuration_periodic_dimensions')
                if frame.ase_config.has('velocities'):
                    push_array_values(jbe, frame.ase_config.get_velocities()*length_libatoms2nomad, 'atom_velocities')
                if frame.ase_config.has('forces'):
                    # TODO Wouldn't it be nicer if forces were added here?
                    pass
                pass

        # SINGLE CONFIGURATIONS
        refs_single_configuration = []
        i_frame = -1
        for frame in all_frames:
            i_frame += 1
            with open_section(jbe, 'section_single_configuration_calculation') as gid:
                refs_single_configuration.append(gid)
                # Reference system description section
                ref_system = refs_system_description[i_frame]
                push_value(jbe, ref_system, 'single_configuration_calculation_to_system_ref')
                # Energy
                if frame.has_energy:
                    push_value(jbe, frame.energy*energy_libatoms2nomad, 'energy_total') # TODO Check units
                # Virial
                if frame.has_virial:
                    push_array_values(jbe, frame.virial, 'x_lib_atoms_virial_tensor')
                # Forces
                if frame.ase_config.has('forces'):
                    push_array_values(jbe, frame.ase_config.get_forces(), 'atom_forces')
                # Type label
                if frame.has_config_type:
                    push_value(jbe, frame.config_type, 'x_lib_atoms_config_type')
                pass

        # FRAME SEQUENCE
        with open_section(jbe, 'section_frame_sequence'):
            push_value(jbe, len(all_frames), 'number_of_frames_in_sequence')
            refs_config = np.array(refs_single_configuration)
            push_array_values(jbe, refs_config, 'frame_sequence_local_frames_ref')

        # GAP DESCRIPTION
        if gap.has_gap_data:
            with open_section(jbe, 'x_lib_atoms_section_gap') as gap_gid:
                push_array_values(jbe, np.array(refs_single_configuration), 'x_lib_atoms_training_config_refs')
                push_value(jbe, gap['GAP_params.label'].As(), 'x_lib_atoms_GAP_params_label')
                push_value(jbe, gap['GAP_params.svn_version'].As(), 'x_lib_atoms_GAP_params_svn_version')
                push_value(jbe, gap['GAP_data.do_core'].As(), 'x_lib_atoms_GAP_data_do_core')
                push_value(jbe, gap['GAP_data.e0'].As(float), 'x_lib_atoms_GAP_data_e0')
                push_value(jbe, gap['command_line.command_line'].As(), 'x_lib_atoms_command_line_command_line')
                push_value(jbe, gap['gpSparse.n_coordinate'].As(int), 'x_lib_atoms_gpSparse_n_coordinate')
                types = [int,         int,          int,              str,          float,             float,         float,             str,     str,                float,   str,          int,                int ]
                keys  = ['n_sparseX', 'dimensions', 'n_permutations', 'sparsified', 'signal_variance', 'signal_mean', 'covariance_type', 'label', 'sparseX_filename', 'theta', 'descriptor', 'perm.permutation', 'perm.i']
                for i,key in enumerate(keys):
                    push_value(jbe, gap['gpCoordinates.%s' % key].As(types[i]), 'x_lib_atoms_gpCoordinates_%s' % key.replace('.', '_'))
                for key in ['alpha', 'sparseX']:
                    push_array_values(jbe, gap['gpCoordinates.%s' % key].As(), 'x_lib_atoms_gpCoordinates_%s' % key.replace('.', '_'))

    jbe.finishedParsingSession("ParseSuccess", None)
    return jbe

if __name__ == '__main__':

    # CALCULATE PATH TO META-INFO FILE
    this_py_file = os.path.abspath(__file__)
    this_py_dirname = os.path.dirname(this_py_file)
    json_supp_file = parser_info["json"]
    meta_info_path = os.path.normpath(os.path.join(this_py_dirname, json_supp_file))

    # LOAD META-INFO FILE
    log("Meta-info from '%s'" % meta_info_path)
    meta_info_env, warns = loadJsonFile(
        filePath=meta_info_path,
        dependencyLoader=None,
        extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS,
        uri=None)

    output_file_name = sys.argv[1]
    parse(output_file_name)











# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the STATPT module from TurboMole"""

import logging
import re
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")


# TODO: include support for transition state search once sample files are provided
class STATPTparser(object):

    def __init__(self, context, key="statpt"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.__trustregion = False
        self.__geo_converged = None

    def purge_data(self):
        self.__trustregion = False
        self.__geo_converged = None

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{15,}[^+ ]+",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(name="Credits",
                    startReStr=r"\s*this is S T A T P T\s*$",
                    coverageIgnore=True,
                    subMatchers=[references],
                    )

        # TODO: add geometry and force extractors for STATPT format
        sub_matchers = [
            self.__context.build_start_time_matcher(),
            header,
            self.build_hessian_update_matcher(),
            self.__context.build_end_time_matcher("statpt")
        ]

        return self.__context.build_module_matcher("statpt", sub_matchers, "STATPT")

    def build_hessian_update_matcher(self):
        def set_trust_region(backend, groups):
            self.__trustregion = True

        def set_hessian_update_method(backend, groups):
            if groups[0] == "BFGS":
                method = "trm_bfgs" if self.__trustregion else "bfgs"
                backend.addValue("geometry_optimization_method", method)

        trust_region_max = SM("\s*Maximum\s+allowed\s+trust\s+radius\s*:\s*"
                              "(?P<x_turbomole_geometry_optimization_trustregion_max__bohr>"
                              + RE_FLOAT+")",
                              name="trust region",
                              startReAction=set_trust_region)
        trust_region_min = SM("\s*Minimum\s+allowed\s+trust\s+radius\s*:\s*"
                              "(?P<x_turbomole_geometry_optimization_trustregion_min__bohr>"
                              + RE_FLOAT+")",
                              name="trust region",
                              startReAction=set_trust_region)
        trust_region_initial = SM("\s*Maximum\s+allowed\s+trust\s+radius\s*:\s*"
                                  "(?P<x_turbomole_geometry_optimization_trustregion_initial__bohr>"
                                  + RE_FLOAT+")",
                                  name="trust region",
                                  startReAction=set_trust_region)
        hessian_update = SM(r"\s*Hessian\s+update\s+method\s*:\s*([^\s].*[^\s]|[^\s])\s*$",
                            name="Hessian update",
                            startReAction=set_hessian_update_method
                            )

        def set_sampling_data(backend, gIndex, section):
            status = all(self.__geo_converged) if self.__geo_converged else None
            self.__context.set_sampling_mode_section(gIndex, status)

        cycle_index = SM(r"\s*ENERGY\s*=\s*"+RE_FLOAT+"\s*a.u.\s*;\s*#\s*of\s*cycle\s*=\s*"
                         r"(?P<x_turbomole_geometry_optimization_cycle_index>[0-9]+)\s*$",
                         name="cycle index")

        return SM(r"\s*\*+\s*Stationary\s*point\s*options\s*\*+\s*$",
                  name="Geo Opt Config",
                  sections=["section_sampling_method"],
                  onClose={"section_sampling_method": set_sampling_data},
                  fixedStartValues={"sampling_method": "geometry optimization"},
                  subMatchers=[
                      trust_region_max,
                      trust_region_min,
                      trust_region_initial,
                      hessian_update,
                      self.build_criteria_matcher(),
                      self.__context["geo"].build_qm_geometry_matcher_statpt(),
                      self.__context["gradient"].build_gradient_matcher_statpt(),
                      cycle_index,
                      self.build_convergence_matcher()
                  ]
                  )

    def build_criteria_matcher(self):
        conv_energy = SM(r"\s*Threshold\s+for\s+energy\s+change\s*:\s*"
                         r"(?P<geometry_optimization_energy_change__hartree>"+RE_FLOAT+")\s*$",
                         name="energy max")
        conv_displace = SM(r"\s*Threshold\s+for\s+max\s+displacement\s+element\s*:\s*"
                           r"(?P<geometry_optimization_geometry_change__bohr>"+RE_FLOAT+")\s*$",
                           name="geometry max")
        conv_force = SM(r"\s*Threshold\s+for\s+max\s+gradient\s+element\s*:\s*"
                        r"(?P<geometry_optimization_threshold_force__forceAu>"
                        + RE_FLOAT+")\s*$",
                        name="force max")
        conv_displace_rms = SM(r"\s*Threshold\s+for\s+RMS\s+of\s+displacement\s*:\s*"
                               r"(?P<x_turbomole_geometry_optimization_geometry_change_rms__bohr>"
                               + RE_FLOAT+")\s*$",
                               name="geometry RMS")
        conv_force_rms = SM(r"\s*Threshold\s+for\s+RMS\s+of\s+gradient\s*:\s*"
                            r"(?P<x_turbomole_geometry_optimization_threshold_force_rms__forceAu>"
                            + RE_FLOAT+")\s*$",
                            name="force RMS")
        return SM(r"\s*\*+\s*Convergence\s+criteria\s*\*+\s*$",
                  name="convergence criteria",
                  subMatchers=[
                      conv_energy,
                      conv_displace,
                      conv_force,
                      conv_displace_rms,
                      conv_force_rms
                  ]
                  )

    def build_convergence_matcher(self):
        def set_initial_convergence(backend, groups):
            self.__geo_converged = list()

        def update_convergence(backend, groups):
            self.__geo_converged.append(groups[0] == "yes")

        def build_check_matcher(criterion):
            return SM(r"\s*"+criterion+"\s+(yes|no)\s+"+RE_FLOAT+"\s+"+RE_FLOAT+"\s*$",
                      name="converged?",
                      startReAction=update_convergence
                      )

        return SM(r"\s*CONVERGENCE\s+INFORMATION\s*$",
                  name="Convergence Test",
                  startReAction=set_initial_convergence,
                  subMatchers=[
                      SM(r"\s*Converged\?\s+Value\s+Criterion\s*$", name="header"),
                      build_check_matcher(r"Energy\s+change"),
                      build_check_matcher(r"RMS\s+of\s+displacement"),
                      build_check_matcher(r"RMS\s+of\s+gradient"),
                      build_check_matcher(r"MAX\s+displacement"),
                      build_check_matcher(r"MAX\s+gradient"),
                  ]
                  )



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
.. autofunc::nomad.utils.create_uuid
.. autofunc::nomad.utils.hash
.. autofunc::nomad.utils.timer

Logging in nomad is structured. Structured logging means that log entries contain
dictionaries with quantities related to respective events. E.g. having the code,
parser, parser version, calc_id, mainfile, etc. for all events that happen during
calculation processing. This means the :func:`get_logger` and all logger functions
take keyword arguments for structured data. Otherwise :func:`get_logger` can
be used similar to the standard *logging.getLogger*.

Depending on the configuration all logs will also be send to a central logstash.

.. autofunc::nomad.utils.get_logger
.. autofunc::nomad.utils.hash
.. autofunc::nomad.utils.create_uuid
.. autofunc::nomad.utils.timer
.. autofunc::nomad.utils.lnr
'''

from typing import List, Iterable
from collections import OrderedDict
import base64
from contextlib import contextmanager
import json
import uuid
import time
import re
import hashlib
import sys
from datetime import timedelta
import collections
import logging

from nomad import config

default_hash_len = 28
''' Length of hashes and hash-based ids (e.g. calc, upload) in nomad. '''

try:
    from . import structlogging
    from .structlogging import legacy_logger

    def get_logger(name, **kwargs):
        '''
        Returns a structlog logger that is already attached with a logstash handler.
        Use additional *kwargs* to pre-bind some values to all events.
        '''
        return structlogging.get_logger(name, **kwargs)

except ImportError:
    def get_logger(name, **kwargs):
        return ClassicLogger(name, **kwargs)


class ClassicLogger:
    '''
    A logger class that emulates the structlog interface, but uses the classical
    build-in Python logging.
    '''
    def __init__(self, name, **kwargs):
        self.kwargs = kwargs
        self.logger = logging.getLogger(name)

    def bind(self, **kwargs):
        all_kwargs = dict(self.kwargs)
        all_kwargs.update(**kwargs)
        return ClassicLogger(self.logger.name, **all_kwargs)

    def __log(self, method_name, event, **kwargs):
        method = getattr(self.logger, method_name)
        all_kwargs = dict(self.kwargs)
        all_kwargs.update(**kwargs)

        message = '%s (%s)' % (
            event,
            ', '.join(['%s=%s' % (str(key), str(value)) for key, value in all_kwargs.items()])
        )
        method(message)

    def __getattr__(self, key):
        return lambda *args, **kwargs: self.__log(key, *args, **kwargs)


def set_console_log_level(level):
    root = logging.getLogger()
    for handler in root.handlers:
        if isinstance(handler, (logging.StreamHandler, logging.FileHandler)):
            handler.setLevel(level)


def decode_handle_id(handle_str: str):
    result = 0
    for c in handle_str:
        ordinal = ord(c.lower())
        if 48 <= ordinal <= 57:
            number = ordinal - 48
        elif 97 <= ordinal <= 118:
            number = ordinal - 87
        else:
            raise ValueError()

        result = result * 32 + number

    return result


def hash(*args, length: int = default_hash_len) -> str:
    ''' Creates a websafe hash of the given length based on the repr of the given arguments. '''
    hash = hashlib.sha512()
    for arg in args:
        hash.update(str(arg).encode('utf-8'))

    return make_websave(hash, length=length)


def make_websave(hash, length: int = default_hash_len) -> str:
    ''' Creates a websafe string for a hashlib hash object. '''
    if length > 0:
        return base64.b64encode(hash.digest(), altchars=b'-_')[:length].decode('utf-8')
    else:
        return base64.b64encode(hash.digest(), altchars=b'-_')[0:-2].decode('utf-8')


def base64_encode(string):
    '''
    Removes any `=` used as padding from the encoded string.
    '''
    encoded = base64.urlsafe_b64encode(string).decode('utf-8')
    return encoded.rstrip("=")


def base64_decode(string):
    '''
    Adds back in the required padding before decoding.
    '''
    padding = 4 - (len(string) % 4)
    bytes = (string + ("=" * padding)).encode('utf-8')
    return base64.urlsafe_b64decode(bytes)


def create_uuid() -> str:
    ''' Returns a web-save base64 encoded random uuid (type 4). '''
    return base64.b64encode(uuid.uuid4().bytes, altchars=b'-_').decode('utf-8')[0:-2]


@contextmanager
def lnr(logger, event, **kwargs):
    '''
    A context manager that Logs aNd Raises all exceptions with the given logger.

    Arguments:
        logger: The logger that should be used for logging exceptions.
        event: the log message
        **kwargs: additional properties for the structured log
    '''
    try:
        yield

    except Exception as e:
        # ignore HTTPException as they are part of the normal flask error handling
        if e.__class__.__name__ == 'HTTPException':
            logger.error(event, exc_info=e, **kwargs)
        raise e


@contextmanager
def timer(logger, event, method='info', **kwargs):
    '''
    A context manager that takes execution time and produces a log entry with said time.

    Arguments:
        logger: The logger that should be used to produce the log entry.
        event: The log message/event.
        method: The log method that should be used. Must be a valid logger method name.
            Default is 'info'.
        **kwargs: Additional logger data that is passed to the log entry.

    Returns:
        The method yields a dictionary that can be used to add further log data.
    '''
    start = time.time()

    try:
        yield kwargs
    finally:
        stop = time.time()

    if logger is None:
        print(event, stop - start)
        return

    logger_method = getattr(logger, 'info', None)
    if logger_method is not None:
        logger_method(event, exec_time=stop - start, **kwargs)
    else:
        logger.error('Unknown logger method %s.' % method)


class archive:
    @staticmethod
    def create(upload_id: str, calc_id: str) -> str:
        return '%s/%s' % (upload_id, calc_id)

    @staticmethod
    def items(archive_id: str) -> List[str]:
        return archive_id.split('/')

    @staticmethod
    def item(archive_id: str, index: int) -> str:
        return archive.items(archive_id)[index]

    @staticmethod
    def calc_id(archive_id: str) -> str:
        return archive.item(archive_id, 1)

    @staticmethod
    def upload_id(archive_id: str) -> str:
        return archive.item(archive_id, 0)


def to_tuple(self, *args):
    return tuple(self[arg] for arg in args)


def chunks(list, n):
    ''' Chunks up the given list into parts of size n. '''
    for i in range(0, len(list), n):
        yield list[i:i + n]


class SleepTimeBackoff:
    '''
    Provides increasingly larger sleeps. Useful when
    observing long running processes with unknown runtime.
    '''

    def __init__(self, start_time: float = 0.1, max_time: float = 5):
        self.current_time = start_time
        self.max_time = max_time

    def __call__(self):
        self.sleep()

    def sleep(self):
        time.sleep(self.current_time)
        self.current_time *= 2
        self.current_time = min(self.max_time, self.current_time)


class ETA:
    def __init__(self, total: int, message: str, interval: int = 1000):
        self.start = time.time()
        self.total = total
        self.count = 0
        self.interval = interval
        self.interval_count = 0
        self.message = message

    def add(self, amount: int = 1):
        self.count += amount
        interval_count = int(self.count / self.interval)
        if interval_count > self.interval_count:
            self.interval_count = interval_count
            delta_t = time.time() - self.start
            eta = delta_t * (self.total - self.count) / self.count
            eta_str = str(timedelta(seconds=eta))
            sys.stdout.write('\r' + (self.message % (self.count, self.total, eta_str)))
            sys.stdout.flush()

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args, **kwargs):
        print('')


def common_prefix(paths):
    '''
    Computes the longest common file path prefix (with respect to '/' separated segments).
    Returns empty string is ne common prefix exists.
    '''
    common_prefix = None

    for path in paths:
        if common_prefix is None:
            common_prefix = path

        index = 0
        index_last_slash = -1
        for a, b in zip(path, common_prefix):
            if a != b:
                break
            if a == '/':
                index_last_slash = index
            index += 1

        if index_last_slash == -1:
            common_prefix = ''
            break

        common_prefix = common_prefix[:index_last_slash + 1]

    if common_prefix is None:
        common_prefix = ''

    return common_prefix


class RestrictedDict(OrderedDict):
    """Dictionary-like container with predefined set of mandatory and optional
    keys and a set of forbidden values.
    """
    def __init__(self, mandatory_keys: Iterable = None, optional_keys: Iterable = None, forbidden_values: Iterable = None, lazy: bool = True):
        """
        Args:
            mandatory_keys: Keys that have to be present.
            optional_keys: Keys that are optional.
            forbidden_values: Values that are forbidden. Only supports hashable values.
            lazy: If false, the values are checked already when inserting. If
                True, the values should be manually checked by calling the
                check()-function.
        """
        super().__init__()

        if isinstance(mandatory_keys, (list, tuple, set)):
            self._mandatory_keys = set(mandatory_keys)
        elif mandatory_keys is None:
            self._mandatory_keys = set()
        else:
            raise ValueError("Please provide the mandatory_keys as a list, tuple or set.")

        if isinstance(optional_keys, (list, tuple, set)):
            self._optional_keys = set(optional_keys)
        elif optional_keys is None:
            self._optional_keys = set()
        else:
            raise ValueError("Please provide the optional_keys as a list, tuple or set.")

        if isinstance(forbidden_values, (list, tuple, set)):
            self._forbidden_values = set(forbidden_values)
        elif forbidden_values is None:
            self._forbidden_values = set()
        else:
            raise ValueError("Please provide the forbidden_values as a list or tuple of values.")

        self._lazy = lazy

    def __setitem__(self, key, value):
        if not self._lazy:

            # Check that only the defined keys are used
            if key not in self._mandatory_keys and key not in self._optional_keys:
                raise KeyError("The key '{}' is not allowed.".format(key))

            # Check that forbidden values are not used.
            try:
                match = value in self._forbidden_values
            except TypeError:
                pass  # Unhashable value will not match
            else:
                if match:
                    raise ValueError("The value '{}' is not allowed.".format(key))

        super().__setitem__(key, value)

    def check(self, recursive=False):
        # Check that only the defined keys are used
        for key in self.keys():
            if key not in self._mandatory_keys and key not in self._optional_keys:
                raise KeyError("The key '{}' is not allowed.".format(key))

        # Check that all mandatory values are all defined
        for key in self._mandatory_keys:
            if key not in self:
                raise KeyError("The mandatory key '{}' is not present.".format(key))

        # Check that forbidden values are not used.
        for key, value in self.items():
            match = False
            try:
                match = value in self._forbidden_values
            except TypeError:
                pass  # Unhashable value will not match
            else:
                if match:
                    raise ValueError("The value '{}' is not allowed but was set for key '{}'.".format(value, key))

        # Check recursively
        if recursive:
            for value in self.values():
                if isinstance(value, RestrictedDict):
                    value.check(recursive)

    def update(self, other):
        for key, value in other.items():
            self.__setitem__(key, value)

    def hash(self) -> str:
        """Creates a hash code from the contents. Ensures consistent ordering.
        """
        hash_str = json.dumps(self, sort_keys=True)

        return hash(hash_str)



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

from __future__ import print_function

# TEMPLATES TAKEN FROM OCTOPUS.  KEEP SYNCHRONIZED.
json_header = """{
  "type": "nomad_meta_info_1_0",
  "description": "autogenerated nomad meta info for %(parser)s parser.  The file which generates this maintained in the parser's git repository.",
  "dependencies": [ {
      "relativePath": "common.nomadmetainfo.json"
    }],
  "metaInfos": [ %(info)s]
}"""


json_template = """{
      "description": "%(description)s",
      "dtypeStr": "%(dtypeStr)s",
      "name": "%(name)s",
      "repeats": false,
      "shape": [],
      "superNames": [
        "%(supername)s"
      ]
    }"""

json_section_template = """{
      "description": "%(description)s",
      "kindStr": "type_abstract_document_content",
      "name": "%(name)s",
      "superNames": [
        "x_siesta_section_input"
      ]
    }, {
      "description": "input section",
      "kindStr": "type_section",
      "name": "x_siesta_section_input",
      "superNames": [
        "section_method"
      ]
    }"""
# END OF TEMPLATES FROM OCTOPUS

import sys
#fname = sys.argv[1]

inputvars_fd = open('inputvars.py', 'w')
print('varlist = [', file=inputvars_fd)

def getlines():
    fd = open('test/H2O/fdf-95611.log')
    for line in fd:
        if not line[:1].isalpha():  # Get rid of blocks
            continue
        yield line

    withinblock = False
    fd = open('test/smeagol-Au-leads/out.fdf')
    for line in fd:
        if line.startswith('%block'):
            withinblock = True
        elif line.startswith('%endblock'):
            withinblock = False
        if withinblock:
            continue
        yield line

seen = set()

#with open(fname) as fd:
#for line in getlines():
if 1:
    varnames = []
    for line in getlines():

        # Get rid of comments
        line = line.strip().rsplit('#', 1)[0].strip()
        if not line:
            continue
        varname = line.split()[0]
        lowervarname = varname.lower()
        if lowervarname in seen:
            continue  # Already registered

        seen.add(lowervarname)
        varnames.append(varname)
    varnames = list(set(varnames))
    varnames.sort()

    jsontokens = []
    for var in varnames:
        print("    '%s'," % var, file=inputvars_fd)
        json = json_template % dict(description=r'siesta input variable \"%s\"'
                                    % var,
                                    name='x_siesta_input_%s' % var,
                                    dtypeStr='C',
                                    supername='x_siesta_input')
        jsontokens.append(json)
    section = json_section_template % dict(description='siesta input '
                                           'variables',
                                           name='x_siesta_input')
    jsontokens.append(section)

    txt = json_header % dict(parser='siesta',
                             info=', '.join(jsontokens))
    print(txt)

print(']', file=inputvars_fd)
